import pandas as pd


def calculate_perc_missing(df):
    """
    Calculate the percentage and total count of missing entries
    in a DataFrame based on a time-index.

    Parameters:
    - df (pd.DataFrame): The input DataFrame with a DatetimeIndex.

    Returns:
    - list: [percentage_missing (float), total_missing (float)]
    """
    # Handle empty DataFrame
    if df.empty:
        raise ValueError("DataFrame is empty")

    # Get the original frequency and reindex with it
    # check if the index is a DatetimeIndex
    if not isinstance(df.index, pd.DatetimeIndex):
        raise ValueError("DataFrame index is not a DatetimeIndex")
    freq = df.index.freq
    if freq is None:
        freq = "1Min"
    # Get the full range of expected dates
    start = df.index.min()
    end = df.index.max()
    expected_index = pd.date_range(start=start, end=end, freq=freq)

    # Calculate missing values
    total_possible = len(expected_index)
    total_actual = len(df)

    total_missing = total_possible - total_actual

    # Calculate percentage
    perc_missing = (total_missing / total_possible) * 100 if total_possible > 0 else 0.0
    perc_missing = round(perc_missing, 2)
    return [perc_missing, 0 if total_missing < 0 else total_missing]



from numpy import nan
from fast_trade.run_backtest import (
    prepare_new_backtest,
    process_logic_and_generate_actions,
    run_backtest,
    take_action,
    clean_field_type,
    process_single_logic,
    process_single_frame,
    determine_action,
    apply_backtest_to_df,
)

from collections import namedtuple
import pandas as pd


def test_take_action_greater_than():
    mock_backtest = [["close", ">", "short"]]
    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    res = take_action(mock_row, mock_backtest, last_frames=[mock_row])
    assert res is True


def test_take_action_less_than():
    mock_backtest = [["close", "<", "short"]]
    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    res = take_action(mock_row, mock_backtest, last_frames=[mock_row])

    assert res is False


def test_take_action_not_equal():
    mock_backtest = [["close", "!=", 0.0219]]
    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    res = take_action(mock_row, mock_backtest, last_frames=[mock_row])
    assert res is True


def test_take_action_no_res():
    # mock_backtest = [["close", "!=", 0.0219]]
    mock_backtest = []
    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    res = take_action(mock_row, mock_backtest, last_frames=[])
    assert res is False


def test_take_action_many_frames():
    mock_backtest = [["close", "=", 0.0212, 2]]
    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    mock_last_frames = [mock_row, mock_row]

    res = take_action(mock_row, mock_backtest, last_frames=mock_last_frames)
    assert res is True


def test_take_action_many_frames_fail():
    mock_backtest = [["close", "=", 0.0212, 3]]
    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row_1 = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    mock_row_2 = MockRow(
        date=1523937966,
        close=0.0219,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    mock_row_3 = MockRow(
        date=1523937999,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    mock_last_frames = [mock_row_1, mock_row_2, mock_row_3]

    res = take_action(mock_row_1, mock_backtest, last_frames=mock_last_frames)

    assert res is False


# WTF
def test_take_action_many_frames_diff():
    mock_backtest = [["close", "=", 0.0212, 3]]
    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row_1 = MockRow(
        date=1523937963,
        close=0.0214,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    mock_row_2 = MockRow(
        date=1523937966,
        close=0.0219,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    mock_row_3 = MockRow(
        date=1523937999,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
    )

    mock_last_frames = [mock_row_1, mock_row_2, mock_row_3]

    res = take_action(mock_row_1, mock_backtest, last_frames=mock_last_frames)

    assert res is False


def test_clean_field_type_num():
    mock_field = "mock_f1"
    MockRow = namedtuple("MockRow", "date close open high low volume short mock_f1")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
        mock_f1=50,
    )

    res = clean_field_type(mock_field, mock_row)

    assert res == 50


def test_clean_field_type_num_str():
    mock_field = "50"
    MockRow = namedtuple("MockRow", "date close open high low volume short mock_f1")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
        mock_f1=50,
    )

    res = clean_field_type(mock_field, mock_row)

    assert res == 50


def test_clean_field_type_float_str():
    mock_field = "50.0"
    MockRow = namedtuple("MockRow", "date close open high low volume short mock_f1")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.01,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
        mock_f1=50,
    )

    res = clean_field_type(mock_field, mock_row)

    assert res == 50


def test_clean_field_type_float():
    mock_field = 50.04
    MockRow = namedtuple("MockRow", "date close open high low volume short mock_f1")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0112,
        mock_f1=50.04,
    )

    res = clean_field_type(mock_field, mock_row)

    assert res == 50.04


def test_clean_field_type_no_row_bool():
    mock_field = True

    res = clean_field_type(mock_field)

    assert res is True


def test_clean_field_type_no_row():
    mock_field = None

    res = clean_field_type(mock_field)

    assert res is None


def test_process_single_logic():
    mock_logic = ["close", "<", "short"]
    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
    )

    res = process_single_logic(mock_logic, mock_row)

    assert res is True


def test_process_single_logic_false():
    mock_logic = ["close", "=", "short"]
    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
    )

    res = process_single_logic(mock_logic, mock_row)

    assert res is False


def test_process_single_logic_false_greater():
    mock_logic = ["close", ">", "short"]
    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
    )

    res = process_single_logic(mock_logic, mock_row)

    assert res is False


def process_single_frame_2(logics, row, require_any):
    results = []

    return_value = False
    for logic in logics:
        res = process_single_logic(logic, row)
        results.append(res)

    if len(results):
        if require_any:
            return_value = any(results)
        else:
            return_value = all(results)

    return return_value


def test_process_single_frame_require_any_false():
    mock_logics = [["close", ">", "short"], ["close", ">", "low"]]

    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
    )
    mock_require_any = False

    res = process_single_frame(mock_logics, mock_row, mock_require_any)

    assert res is False


def test_process_single_frame_require_any_true():
    mock_logics = [["close", ">", "short"], ["close", ">", "low"]]

    MockRow = namedtuple("MockRow", "date close open high low volume short")
    mock_row = MockRow(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
    )
    mock_require_any = True

    res = process_single_frame(mock_logics, mock_row, mock_require_any)

    assert res is True


def test_determine_action_1():
    MockFrame = namedtuple("MockFrame", "date close open high low volume short")
    mock_frame = MockFrame(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
    )

    mock_backtest = {
        "exit": [["close", "<", "short"]],
        "any_exit": [],
        "enter": [],
        "any_enter": [],
    }

    res = determine_action(mock_frame, mock_backtest, last_frames=[mock_frame])

    assert res == "x"


def test_determine_action_2():
    MockFrame = namedtuple("MockFrame", "date close open high low volume short")
    mock_frame = MockFrame(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
    )

    mock_backtest = {
        "exit": [["close", ">", "short"]],
        "any_exit": [],
        "enter": [],
        "any_enter": [],
    }

    res = determine_action(mock_frame, mock_backtest, last_frames=[mock_frame])

    assert res == "h"


def test_determine_action_2_any_exit():
    MockFrame = namedtuple("MockFrame", "date close open high low volume short")
    mock_frame = MockFrame(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
    )

    mock_backtest = {
        "exit": [],
        "any_exit": [["close", "<", "short"]],
        "enter": [],
        "any_enter": [],
    }

    res = determine_action(mock_frame, mock_backtest, last_frames=[mock_frame])

    assert res == "ax"


def test_determine_action_3_any_exit():
    MockFrame = namedtuple("MockFrame", "date close open high low volume short")
    mock_frame = MockFrame(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
    )

    mock_backtest = {
        "exit": [["close", ">", "short"]],
        "any_exit": [["close", "<", "short"]],
        "enter": [],
        "any_enter": [],
    }

    res = determine_action(mock_frame, mock_backtest, last_frames=[mock_frame])

    assert res == "ax"


def test_determine_action_1_trailing_stop_loss():
    MockFrame = namedtuple(
        "MockFrame", "date close open high low volume short trailing_stop_loss"
    )
    mock_frame = MockFrame(
        date=1523937963,
        close=0.0212,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
        trailing_stop_loss=10,
    )

    mock_backtest = {
        "exit": [["close", ">", "short"]],
        "any_exit": [["close", ">", "short"]],
        "enter": [],
        "any_enter": [],
        "trailing_stop_loss": 0.0213,
    }

    res = determine_action(mock_frame, mock_backtest, last_frames=[mock_frame])

    assert res == "tsl"


def test_determine_action_enter_1_mult():
    MockFrame = namedtuple("MockFrame", "date close open high low volume short")
    mock_frame = MockFrame(
        date=1523937963,
        close=0.0588,
        open=0.0133,
        high=0.125,
        low=0.01,
        volume=36898,
        short=0.0312,
    )

    mock_backtest = {
        "exit": [],
        "any_exit": [],
        "enter": [["close", ">", "short"], ["high", ">", "close"]],
        "any_enter": [],
    }

    res = determine_action(mock_frame, mock_backtest, last_frames=[mock_frame])

    assert res == "e"


def test_determine_action_any_enter():
    MockFrame = namedtuple("MockFrame", "date close open high low volume short")
    mock_frame = MockFrame(
        date=1523937963,
        close=0.0588,
        open=0.0133,
        high=0.025,
        low=0.01,
        volume=36898,
        short=0.0312,
    )

    mock_backtest = {
        "exit": [],
        "any_exit": [],
        "enter": [["close", ">", "19"]],
        "any_enter": [["close", ">", "short"], ["close", ">", 100]],
    }

    res = determine_action(mock_frame, mock_backtest, last_frames=[mock_frame])

    assert res == "ae"


def test_proccess_logic_and_actions_no_logics():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    mock_backtest = {
        "exit": [],
        "any_exit": [],
        "enter": [],
        "any_enter": [],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)

    assert "h" in res.action.unique()


def test_proccess_logic_and_actions_single_enter():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    mock_backtest = {
        "exit": [],
        "any_exit": [],
        "enter": [["volume", ">", 171000]],
        "any_enter": [],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)
    assert list(res.action.values) == ["h", "h", "h", "h", "h", "e", "e", "e", "e"]


def test_proccess_logic_and_actions_multi_enter():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    # fake an indicator
    mock_df["ind_1"] = [0, 1, 2, 4, 5, 6, 7, 8, 9]
    mock_backtest = {
        "exit": [],
        "any_exit": [],
        "enter": [["ind_1", "<", 9], ["ind_1", ">", 7]],
        "any_enter": [],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)

    assert list(res.action.values) == [
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
        "e",
        "h",
    ]


def test_proccess_logic_and_actions_single_any_enter():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    # fake an indicator
    mock_df["ind_1"] = [0, 1, 2, 4, 5, 6, 7, 8, 9]
    mock_backtest = {
        "exit": [],
        "any_exit": [],
        "enter": [],
        "any_enter": [["ind_1", ">", 8]],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)
    assert list(res.action.values) == [
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
        "ae",
    ]


def test_proccess_logic_and_actions_single_exit():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    # fake an indicator
    mock_df["ind_1"] = [0, 1, 2, 4, 5, 6, 7, 8, 9]
    mock_backtest = {
        "exit": [["ind_1", ">", 7]],
        "any_exit": [],
        "enter": [],
        "any_enter": [],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)
    assert list(res.action.values) == [
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
        "x",
        "x",
    ]


def test_proccess_logic_and_actions_single_any_exit():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    mock_df["ind_1"] = [0, 0, 0, 0, 1, 1, 1, 0, 0]
    mock_backtest = {
        "exit": [],
        "any_exit": [["ind_1", "=", 1]],
        "enter": [],
        "any_enter": [],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)
    assert list(res.action) == [
        "h",
        "h",
        "h",
        "h",
        "ax",
        "ax",
        "ax",
        "h",
        "h",
    ]


def test_proccess_logic_and_actions_exit_enter():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    # fake an indicator
    mock_df["ind_1"] = [0, 1, 2, 4, 5, 5, 4, 4, 3]
    mock_backtest = {
        "exit": [["ind_1", "=", 4]],
        "any_exit": [],
        "enter": [["ind_1", "=", 5]],
        "any_enter": [],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)
    assert list(res.action.values) == [
        "h",
        "h",
        "h",
        "x",
        "e",
        "e",
        "x",
        "x",
        "h",
    ]


def test_proccess_logic_and_actions_multi_enter_exit():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    # fake an indicator
    mock_df["ind_1"] = [0, 1, 2, 4, 5, 6, 7, 8, 9]
    mock_backtest = {
        "exit": [["ind_1", ">", 8], ["ind_1", "<", 10]],
        "any_exit": [],
        "enter": [["ind_1", "<", 9], ["ind_1", ">", 7]],
        "any_enter": [],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)

    assert list(res.action.values) == [
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
        "e",
        "x",
    ]


def test_proccess_logic_and_actions_enter_exit_confirmations():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    # fake an indicator
    mock_df["ind_1"] = [0, 1, 2, 4, 5, 6, 7, 8, 9]
    mock_df["ind_2"] = [5, 5, 4, 2, 6, 4, 9, 9, 1]
    mock_backtest = {
        "enter": [["ind_1", "<", 2]],
        "any_enter": [],
        "exit": [["ind_2", ">", 8, 2]],
        "any_exit": [],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)
    print(res)

    assert list(res.action.values) == [
        "e",
        "e",
        "h",
        "h",
        "h",
        "h",
        "h",
        "x",
        "h",
    ]


def test_proccess_logic_and_actions_enter_exit_confirmations_multi():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    # fake an indicator
    mock_df["ind_1"] = [0, 1, 2, 4, 5, 6, 7, 8, 9]
    mock_df["ind_2"] = [5, 5, 4, 2, 6, 4, 9, 9, 1]
    mock_df["ind_3"] = [1, 1, 2, 3, 5, 9, 9, 9, 8]

    mock_backtest = {
        "enter": [["ind_1", "<", 2, 2]],
        "any_enter": [],
        "exit": [["ind_2", ">", 8, 2], ["ind_3", "<", 8, 3]],
        "any_exit": [],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)

    assert list(res.action.values) == [
        "h",
        "e",
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
        "h",
    ]


def test_proccess_logic_and_actions_enter_exit_confirmations_multi_2():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    # fake an indicator
    mock_df["ind_1"] = [5, 5, 5, 2, 6, 7, 9, 9, 1]

    mock_backtest = {
        "enter": [["ind_1", "=", 5, 2]],
        "any_enter": [],
        "exit": [["ind_1", ">", 6, 3]],
        "any_exit": [],
    }

    res = process_logic_and_generate_actions(mock_df, mock_backtest)

    print(res)
    assert list(res.action.values) == [
        "h",
        "e",
        "e",
        "h",
        "h",
        "h",
        "h",
        "x",
        "h",
    ]


def test_prepare_new_backtest_simple():
    mock_backtest = {"base_balance": 126, "exit_on_end": False, "comission": 0.022}

    res = prepare_new_backtest(mock_backtest)

    assert res["exit_on_end"] is False
    assert res["base_balance"] == 126
    assert res["comission"] != 0


def test_apply_backtest_to_df():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    # fake an indicator
    mock_df["ind_1"] = [5, 5, 5, 2, 6, 7, 9, 9, 1]

    mock_backtest = {
        "base_balance": 1000,
        "exit_on_end": False,
        "comission": 0.01,
        "lot_size_perc": 1,
        "enter": [["ind_1", "=", 5, 2]],
        "any_enter": [],
        "exit": [["ind_1", ">", 6, 3]],
        "any_exit": [],
    }

    res = apply_backtest_to_df(mock_df, mock_backtest)

    assert "adj_account_value_change_perc" in list(res.columns)
    assert "adj_account_value_change" in list(res.columns)



import datetime
import pandas as pd
import re
import itertools

from .run_analysis import apply_logic_to_df
from .build_summary import build_summary
from fast_trade.archive.db_helpers import get_kline
from .build_data_frame import prepare_df
from .validate_backtest import validate_backtest, validate_backtest_with_df
from .evaluate import evaluate_rules


class MissingData(Exception):
    pass


class MissingBacktestKeyException(Exception):
    def __init__(self, missing_keys):
        self.missing_keys = "\n".join(missing_keys)
        super().__init__(f"Missing keys: {self.missing_keys}")


def run_backtest(backtest: dict, df: pd.DataFrame = pd.DataFrame(), summary=True):
    """
    Run a backtest on a given dataframe
    Parameters
        backtest: dict, required, object containing the logic to test and other details
        data_path: string or list, required, where to find the csv file of the ohlcv data
        df: pandas dataframe indexed by date
    Returns
        dict
            summary dict, summary of the performace of backtest
            df dataframe, object used in the backtest
            trade_log, dataframe of all the rows where transactions happened
    """

    performance_start_time = datetime.datetime.utcnow()
    new_backtest = prepare_new_backtest(backtest)
    errors = validate_backtest(new_backtest)

    if errors.get("has_error"):
        # find all the keys with values
        error_keys = [
            key for key, value in errors.items() if value and key != "has_error"
        ]

        raise MissingBacktestKeyException(error_keys)

    if df.empty:
        # check the local archive for the data
        # calculate the start and end dates based on the max number of periods in any dp args
        def get_max_periods(datapoint):
            args = datapoint.get("args", [])
            periods = [int(arg) for arg in args if isinstance(arg, int)]
            return max(periods)

        args = [get_max_periods(dp) for dp in new_backtest.get("datapoints", [])]
        max_periods = max(args)
        # print(max_periods)
        # get the frequency of the backtest
        freq = new_backtest.get("freq")
        if not freq and new_backtest.get("chart_period"):
            freq = new_backtest.get("chart_period")
        # convert the frequency to a timedelta
        td_freq = pd.Timedelta(freq)

        start_date = backtest.get("start_date", None)
        if start_date:
            start_date = datetime.datetime.fromisoformat(start_date)
            start_date = start_date - td_freq * max_periods

        # get the data from the local archive
        df = get_kline(
            backtest.get("symbol"),
            backtest.get("exchange"),
            start_date,
            backtest.get("end_date"),
            freq=backtest.get("freq") or backtest.get("chart_period"),
        )

    if df.empty:
        raise MissingData(
            f"No data found for {backtest.get('symbol')} on {backtest.get('exchange')} or in the given dataframe"
        )

    df = prepare_df(df, new_backtest)

    df = apply_backtest_to_df(df, new_backtest)
    # throw an error if the backtest is not valid
    validate_backtest_with_df(new_backtest, df)

    if summary:
        summary, trade_log = build_summary(df, performance_start_time)
    else:
        performance_stop_time = datetime.datetime.utcnow()
        summary = {
            "test_duration": (
                performance_stop_time - performance_start_time
            ).total_seconds()
        }
        trade_log = pd.DataFrame()

    rule_eval = evaluate_rules(summary, new_backtest.get("rules", []))
    summary["rules"] = {
        "all": rule_eval[0],
        "any": rule_eval[1],
        "results": rule_eval[2],
    }
    # add the strategy to the summary
    summary["strategy"] = new_backtest
    return {
        "summary": summary,
        "df": df,
        "trade_df": trade_log,
        "backtest": new_backtest,
    }


def prepare_new_backtest(backtest):
    """
    Parameters
    ----------
        backtest, a raw backest object

    Returns
    -------
        backtest, a backtest dictionary normalized with defaults

    """
    new_backtest = backtest.copy()

    new_backtest["base_balance"] = backtest.get("base_balance", 1000)
    new_backtest["exit_on_end"] = backtest.get("exit_on_end", False)
    new_backtest["comission"] = backtest.get("comission", 0)
    new_backtest["trailing_stop_loss"] = backtest.get("trailing_stop_loss", 0)
    # new_backtest["any_enter"] = backtest.get("any_enter", [])
    # new_backtest["any_exit"] = backtest.get("any_exit", [])
    new_backtest["lot_size_perc"] = float(backtest.get("lot_size", 1))
    new_backtest["max_lot_size"] = int(backtest.get("max_lot_size", 0))
    new_backtest["rules"] = backtest.get("rules", [])

    # if chart_start and chart_stop are provided, use them
    if backtest.get("chart_start"):
        new_backtest["start"] = backtest.get("chart_start")
        del new_backtest["chart_start"]
        print("Warning: chart_start is deprecated, use start instead.")
    if backtest.get("chart_stop"):
        new_backtest["stop"] = backtest.get("chart_stop")
        del new_backtest["chart_stop"]
        print("Warning: chart_stop is deprecated, use stop instead.")

    return new_backtest


def apply_backtest_to_df(df: pd.DataFrame, backtest: dict):
    """Processes the frame and adds the resultent rows
    Parameters
    ----------
        df, dataframe with all the calculated datapoints
        backtest, backtest object

    Returns
    -------
        df, dataframe with with all the actions and backtest processed
    """

    df = process_logic_and_generate_actions(df, backtest)

    df = apply_logic_to_df(df, backtest)

    df["adj_account_value_change_perc"] = df["adj_account_value"].pct_change()
    df["adj_account_value_change"] = df["adj_account_value"].diff()

    # set the index to the date
    df.index = pd.to_datetime(df.index)
    df.index.name = "date"
    return df


def process_logic_and_generate_actions(df: pd.DataFrame, backtest: object):
    """
    Parameters
    ----------
        df, dataframe with the datapoints (indicators) calculated
        backtest, backtest object

    Returns
    -------
        df, a modified dataframe with the "actions" added

    Explainer
    ---------
    In this function, like the name suggests, we process the logic and generate the actions.

    """

    """we need to search though all the logics and find the highest confirmation number
    so we know how many frames to pass in
    """
    logics = [
        backtest.get("enter", []),
        backtest.get("exit", []),
        backtest.get("any_exit", []),
        backtest.get("any_enter", []),
    ]

    logics = list(itertools.chain(*logics))
    max_last_frames = 0  # the we need to keep the
    for logic in logics:
        if len(logic) > 3:
            if logic[3] > max_last_frames:
                max_last_frames = logic[3]

    if max_last_frames:
        actions = []
        last_frames = []
        for frame in df.itertuples():
            last_frames.insert(0, frame)
            if len(last_frames) >= max_last_frames + 1:
                last_frames.pop()
            wtf = determine_action(frame, backtest, last_frames)
            actions.append(wtf)
        df["action"] = actions
    else:
        df["action"] = [determine_action(frame, backtest) for frame in df.itertuples()]

    return df


def determine_action(frame: pd.DataFrame, backtest: dict, last_frames=[]):
    """processes the actions with the applied logic
    Parameters
    ----------
        frame: current row of the dataframe
        backtest: object with the logic of how to trade

    Returns
    -------
        string, "e" (enter), "x" (exit), "h" (hold) of what
        the backtest would do
    """

    trailing_stop_loss = backtest.get("trailing_stop_loss")

    if trailing_stop_loss:
        if frame.close <= frame.trailing_stop_loss:
            return "tsl"

    if take_action(frame, backtest.get("exit", []), last_frames):
        return "x"

    if take_action(frame, backtest.get("any_exit", []), last_frames, require_any=True):
        return "ax"

    if take_action(frame, backtest.get("enter", []), last_frames):
        return "e"

    if take_action(frame, backtest.get("any_enter", []), last_frames, require_any=True):
        return "ae"

    return "h"


def take_action(current_frame, logics, last_frames=[], require_any=False):
    """determines whether to take action based on the logic in the backtest
    Parameters
    ----------
        row: data row to operate on
        backtest: dictionary of logic and how to impliment it

    Returns
    -------
        boolean, True if row meets the criteria of given backtest,
        False if otherwise
    """

    results = []

    if len(last_frames):
        for logic in logics:
            if len(logic) > 3:
                frames = logic[3]
                lcl_frames = last_frames[:frames]
                lcl_results = []

                if len(lcl_frames) < frames:
                    res = False
                else:
                    for lcl_frame in lcl_frames:
                        lcl_res = process_single_frame(
                            [logic], lcl_frame, require_any=False
                        )
                        lcl_results.append(lcl_res)
                    res = all(lcl_results)
                results.append(res)
            else:
                res = process_single_frame([logic], current_frame, require_any)
                results.append(res)
    else:
        res = process_single_frame(logics, current_frame, require_any)
        results.append(res)

    ret_value = False
    if len(results):
        if require_any:
            ret_value = any(results)
        else:
            ret_value = all(results)
    return ret_value


def process_single_frame(logics, row, require_any):
    results = []

    return_value = False
    for logic in logics:
        res = process_single_logic(logic, row)
        results.append(res)

    if len(results):
        if require_any:
            return_value = any(results)
        else:
            return_value = all(results)

    return return_value


def process_single_logic(logic, row):
    val0 = clean_field_type(logic[0], row=row)
    val1 = clean_field_type(logic[2], row=row)

    if logic[1] == ">":
        return_value = bool(val0 > val1)
    if logic[1] == "<":
        return_value = bool(val0 < val1)
    if logic[1] == "=":
        return_value = bool(val0 == val1)
    if logic[1] == "!=":
        return_value = bool(val0 != val1)
    if logic[1] == ">=":
        return_value = bool(val0 >= val1)
    if logic[1] == "<=":
        return_value = bool(val0 <= val1)

    return return_value


def clean_field_type(field, row=None):
    """Determines the value of what to run the logic against.
        This might be a calculated value from the current row,
        or a supplied value, such as a number.

    Parameters
    ----------
        field - str, int, or float, logic field to check
        row - dict, dictionary of values of the current frame

    Returns
    -------
        str or int

    """
    if row:
        if not isinstance(row, dict):
            row = row._asdict()

    if isinstance(field, str):
        if field.isnumeric():
            return int(field)
        if re.match(r"^-?\d+(?:\.\d+)$", field):  # if its a string in a float
            return float(field)

    if type(field) is bool:
        return field

    if isinstance(field, int) or isinstance(field, float):

        return field

    if row:
        return row[field]

    return row



# flake8: noqa
import json
import os
import datetime
import pandas as pd
import matplotlib.pyplot as plt
import requests
import re
from fast_trade.archive.db_helpers import connect_to_db

ARCHIVE_PATH = os.getenv("ARCHIVE_PATH", "ft_archive")


class MissingStrategyFile(Exception):
    pass


def open_strat_file(fp):
    reg = r"https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*)"

    is_url = re.search(reg, fp)
    if is_url:
        # url
        req = requests.get(fp)
        if req.status_code in [200, 201, 202, 301]:
            return req.json()
        else:
            raise MissingStrategyFile(
                "Could not open strategy file at url: {}".format(fp)
            )

    strat_obj = {}
    try:
        with open(fp, "r") as json_file:
            strat_obj = json.load(json_file)
            return strat_obj

    except FileNotFoundError:
        raise MissingStrategyFile("Could not open strategy file at path: {}".format(fp))


def create_plot(df, trade_df):
    # Filter for numeric columns only
    numeric_df = df.select_dtypes(include=["number"])

    numeric_df = numeric_df.drop(
        columns=[
            "open",
            "high",
            "low",
            "adj_account_value",
            "account_value",
            "aux",
            "fee",
            "adj_account_value_change_perc",
            "adj_account_value_change",
        ]
    )

    # Calculate the range of each column
    ranges = numeric_df.max() - numeric_df.min()

    # Define a threshold for grouping columns with similar ranges
    close_range = ranges["close"]
    threshold = close_range * 0.1  # Adjust this factor as needed

    # Group columns based on their range
    close_group = [
        col for col in numeric_df.columns if abs(ranges[col] - close_range) < threshold
    ]
    separate_group = [col for col in numeric_df.columns if col not in close_group]

    # Create subplots
    num_plots = 1 + len(
        separate_group
    )  # One plot for the close group, others for separate columns
    fig, axs = plt.subplots(num_plots, 1, figsize=(10, 6 * num_plots), sharex=True)

    # Ensure axs is iterable
    if num_plots == 1:
        axs = [axs]

    # Plot the close group together
    for column in close_group:
        numeric_df.plot(y=column, ax=axs[0], legend=True, label=column)

    # Iterate over the trade DataFrame and plot each point with the appropriate color
    for index, row in trade_df.iterrows():
        color = "green" if row["in_trade"] else "red"
        for column in close_group:
            axs[0].scatter(index, row[column], color=color, s=10, alpha=0.7)

    axs[0].set_title("Close Group Over Time")
    axs[0].set_ylabel("Value")
    axs[0].grid(True, linestyle="--", alpha=0.5)

    # Plot each separate column in its own subplot
    for ax, column in zip(axs[1:], separate_group):
        numeric_df.plot(y=column, ax=ax, legend=False, label=column)

        for index, row in trade_df.iterrows():
            color = "green" if row["in_trade"] else "red"
            ax.scatter(index, row[column], color=color, s=10, alpha=0.7)

        ax.set_title(f"{column} Over Time")
        ax.set_ylabel(column)
        ax.grid(True, linestyle="--", alpha=0.5)

    # Set the x-axis label for the last subplot
    axs[-1].set_xlabel("Date")

    # Adjust layout
    plt.tight_layout()
    plt.show()
    return fig


def save(result):
    """
    Save the dataframe, backtest, and plot into the specified path
    """

    save_path = ARCHIVE_PATH
    if not os.path.exists(save_path):
        os.mkdir(save_path)
    if not os.path.exists(f"{save_path}/backtests"):
        os.mkdir(f"{save_path}/backtests")
    # dir exists, now make a new dir with the files
    new_dir = (
        f"{datetime.datetime.strftime(datetime.datetime.now(), '%Y_%m_%d_%H_%M_%S')}"
    )

    new_save_dir = f"{save_path}/backtests/{new_dir}"

    os.mkdir(new_save_dir)

    # save the backtest args
    # summary file
    with open(f"{new_save_dir}/summary.json", "w") as summary_file:
        summary_file.write(json.dumps(result["summary"], indent=2))

    # dataframe
    # result["df"].to_csv(f"{new_save_dir}/dataframe.csv")
    # result["trade_df"].to_csv(f"{new_save_dir}/trade_dataframe.csv")
    df_con = connect_to_db(f"{new_save_dir}/dataframe.db", create=True)
    result["df"].to_sql(
        "dataframe", con=df_con, if_exists="replace", index=True, index_label="date"
    )
    trade_con = connect_to_db(f"{new_save_dir}/trade_log.db", create=True)
    result["trade_df"].to_sql(
        "trade_log", con=trade_con, if_exists="replace", index=True, index_label="date"
    )

    # plot
    create_plot(result["df"], result["trade_df"])

    plt.savefig(f"{new_save_dir}/plot.png")



import pandas as pd


def to_dataframe(ticks: list) -> pd.DataFrame:
    """Convert list to Series compatible with the library."""

    df = pd.DataFrame(ticks)
    df["time"] = pd.to_datetime(df["time"], unit="s")
    df.set_index("time", inplace=True)

    return df


def resample(df: pd.DataFrame, interval: str) -> pd.DataFrame:
    """Resample DataFrame by <interval>."""

    d = {"open": "first", "high": "max", "low": "min", "close": "last", "volume": "sum"}

    return df.resample(interval).agg(d)


def resample_calendar(df: pd.DataFrame, offset: str) -> pd.DataFrame:
    """Resample the DataFrame by calendar offset.
    See http://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#anchored-offsets for compatible offsets.
    :param df: data
    :param offset: calendar offset
    :return: result DataFrame
    """

    d = {"open": "first", "high": "max", "low": "min", "close": "last", "volume": "sum"}

    return df.resample(offset).agg(d)


def trending_up(df: pd.Series, period: int) -> pd.Series:
    """returns boolean Series if the inputs Series is trending up over last n periods.
    :param df: data
    :param period: range
    :return: result Series
    """

    return pd.Series(df.diff(period) > 0, name="trending_up {}".format(period))


def trending_down(df: pd.Series, period: int) -> pd.Series:
    """returns boolean Series if the input Series is trending up over last n periods.
    :param df: data
    :param period: range
    :return: result Series
    """

    return pd.Series(df.diff(period) < 0, name="trending_down {}".format(period))


def infer_frequency(df: pd.DataFrame) -> str:
    """Infers the frequency of a DataFrame by analyzing time differences between consecutive index values.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with a datetime index

    Returns
    -------
    str
        The inferred frequency as a string (e.g., '1Min', '5Min', '1H', '1D', etc.)
        Returns None if frequency cannot be determined
    """
    if not isinstance(df.index, pd.DatetimeIndex):
        raise ValueError("DataFrame index must be a DatetimeIndex")

    if df.index.freq is not None:
        return df.index.freq

    # Calculate time differences between consecutive index values
    time_diffs = df.index.to_series().diff()

    # Get the most common time difference
    most_common_diff = time_diffs.mode()[0]
    seconds = most_common_diff.total_seconds()

    # Convert seconds to appropriate frequency string
    if seconds < 60:
        return f"{int(seconds)}S"
    elif seconds < 3600:
        minutes = int(seconds / 60)
        return f"{minutes}Min"
    elif seconds < 86400:
        hours = int(seconds / 3600)
        return f"{hours}H"
    else:
        days = int(seconds / 86400)
        return f"{days}D"



import pytest
import pandas as pd
import random
from fast_trade.run_analysis import (
    calculate_new_account_value_on_enter,
    convert_base_to_aux,
    convert_aux_to_base,
    apply_logic_to_df,
    enter_position,
    exit_position,
    calculate_fee,
)


def test_convert_base_to_aux_1():
    mock_close = 10
    mock_last_base = 100
    res = convert_base_to_aux(mock_last_base, mock_close)

    assert res == 10


def test_convert_base_to_aux_2():
    mock_close = 0.025
    mock_last_base = 60

    res = convert_base_to_aux(mock_last_base, mock_close)
    assert res == 2400.0


def test_convert_base_to_aux_3():
    mock_close = 0.3123
    mock_last_base = 212.2333

    res = convert_base_to_aux(mock_last_base, mock_close)
    assert res == 679.58149215


def test_convert_base_to_aux_4():
    mock_close = 0.3123
    mock_last_base = 0

    res = convert_base_to_aux(mock_last_base, mock_close)
    assert res == 0


def test_convert_base_to_aux_0s():
    mock_close = 0
    mock_last_base = 0

    res = convert_base_to_aux(mock_last_base, mock_close)
    assert res == 0


def test_convert_aux_to_base_1():
    mock_close = 0.99992
    mock_last_aux = 133.22

    aux = convert_aux_to_base(mock_last_aux, mock_close)

    assert aux == 133.2093424


def test_convert_aux_to_base_2():
    mock_close = 0.99992
    mock_base_balance = 0

    aux = convert_aux_to_base(mock_close, mock_base_balance)

    assert aux == 0.0


def test_calculate_fee():
    mock_price = 100
    mock_comission = 0.01

    fee = calculate_fee(mock_price, mock_comission)

    assert fee == 0.01


def test_calculate_fee_again():
    mock_price = 200
    mock_comission = 0.1

    fee = calculate_fee(mock_price, mock_comission)

    assert fee == 0.2


def test_calculate_fee_larger():
    mock_order_size = 1000
    mock_comission = 0.1

    fee = calculate_fee(mock_order_size, mock_comission)

    assert fee == 1


def test_calculate_fee_no_comission():
    mock_price = 100
    mock_comission = 0.00

    fee = calculate_fee(mock_price, mock_comission)

    assert fee == 0.0


def test_enter_position_1():
    mock_account_value_list = []
    mock_lot_size = 1
    mock_account_value = 1000
    mock_max_lot_size = 0
    mock_close = 10
    mock_comission = 0

    in_trade, new_aux, new_account_value, fee = enter_position(
        mock_account_value_list,
        mock_lot_size,
        mock_account_value,
        mock_max_lot_size,
        mock_close,
        mock_comission,
    )

    assert in_trade is True
    assert new_aux == 100.0
    assert fee == 0.0
    assert new_account_value == 0


def test_enter_position_2():
    mock_account_value_list = [1000]
    mock_account_value = 0
    mock_lot_size = 1
    mock_max_lot_size = 0
    mock_close = 10
    mock_comission = 0

    in_trade, new_aux, new_account_value, fee = enter_position(
        mock_account_value_list,
        mock_lot_size,
        mock_account_value,
        mock_max_lot_size,
        mock_close,
        mock_comission,
    )

    assert in_trade is True
    assert new_aux == 100.0
    assert fee == 0.0
    assert new_account_value == 0


def test_enter_position_3():
    mock_account_value_list = [1000, 0, 1100]
    mock_account_value = 1100
    mock_lot_size = 1
    mock_max_lot_size = 0
    mock_close = 10
    mock_comission = 0

    in_trade, new_aux, new_account_value, fee = enter_position(
        mock_account_value_list,
        mock_lot_size,
        mock_account_value,
        mock_max_lot_size,
        mock_close,
        mock_comission,
    )

    assert in_trade is True
    assert new_aux == 110.0
    assert fee == 0.0
    assert new_account_value == 0


def test_enter_position_lot_size():
    mock_account_value_list = [1000, 0, 1100]
    mock_account_value = 1100
    mock_lot_size = 0.5
    mock_max_lot_size = 0
    mock_close = 10
    mock_comission = 0

    in_trade, new_aux, new_account_value, fee = enter_position(
        mock_account_value_list,
        mock_lot_size,
        mock_account_value,
        mock_max_lot_size,
        mock_close,
        mock_comission,
    )

    assert in_trade is True
    assert new_aux == 55.0
    assert fee == 0.0
    assert new_account_value == 550.0


def test_enter_position_comission():
    mock_account_value_list = []
    mock_lot_size = 1
    mock_account_value = 1000
    mock_max_lot_size = 0
    mock_close = 10
    mock_comission = 0.01

    in_trade, new_aux, new_account_value, fee = enter_position(
        mock_account_value_list,
        mock_lot_size,
        mock_account_value,
        mock_max_lot_size,
        mock_close,
        mock_comission,
    )

    assert in_trade is True
    assert new_aux == 99.99

    assert fee == 0.01
    assert new_account_value == 0


def test_enter_position_comission_and_lot_size():
    mock_account_value_list = []
    mock_lot_size = 0.5
    mock_account_value = 1000
    mock_max_lot_size = 0
    mock_close = 10
    mock_comission = 0.01

    in_trade, new_aux, new_account_value, fee = enter_position(
        mock_account_value_list,
        mock_lot_size,
        mock_account_value,
        mock_max_lot_size,
        mock_close,
        mock_comission,
    )

    assert in_trade is True
    assert new_aux == 49.995
    assert fee == 0.005
    assert new_account_value == 500


def test_enter_position_max_lot_size():
    mock_account_value_list = []
    mock_lot_size = 0.5
    mock_account_value = 1000
    mock_max_lot_size = 100
    mock_close = 10
    mock_comission = 0.01

    in_trade, new_aux, new_account_value, fee = enter_position(
        mock_account_value_list,
        mock_lot_size,
        mock_account_value,
        mock_max_lot_size,
        mock_close,
        mock_comission,
    )

    assert in_trade is True
    assert new_aux == 9.999
    assert fee == 0.001
    assert new_account_value == 900


def test_exit_position_basic():
    mock_account_value_list = [1000, 0]
    mock_aux = 100
    mock_close = 11
    mock_comission = 0

    in_trade, new_aux, new_account_value, fee = exit_position(
        mock_account_value_list, mock_close, mock_aux, mock_comission
    )

    assert in_trade is False
    assert new_aux == 0
    assert fee == 0.0
    assert new_account_value == 1100


def test_exit_position_without_account_value():
    # There should always be an account available if we are exiting the trade.
    mock_account_value_list = []
    mock_aux = 100
    mock_close = 11
    mock_comission = 0

    with pytest.raises(IndexError):
        exit_position(mock_account_value_list, mock_close, mock_aux, mock_comission)


def test_exit_position_as_second_tick():
    mock_account_value_list = [500]
    mock_aux = 100
    mock_close = 11
    mock_comission = 0

    in_trade, new_aux, new_account_value, fee = exit_position(
        mock_account_value_list, mock_close, mock_aux, mock_comission
    )

    assert in_trade is False
    assert new_aux == 0
    assert fee == 0.0
    assert new_account_value == 1600


def test_calculate_new_account_value_on_enter_basic():
    mock_base_transaction_amount = 1000
    mock_account_value_list = []
    mock_account_value = 1000

    new_account_value = calculate_new_account_value_on_enter(
        mock_base_transaction_amount, mock_account_value_list, mock_account_value
    )
    assert new_account_value == 0


def test_calculate_new_account_value_on_enter_with_account_vaue_list():
    mock_base_transaction_amount = 600
    mock_account_value_list = [1000]
    mock_account_value = 1000

    new_account_value = calculate_new_account_value_on_enter(
        mock_base_transaction_amount, mock_account_value_list, mock_account_value
    )

    assert new_account_value == 400


def test_apply_logic_to_df_simple():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )

    mock_df.index = pd.to_datetime(mock_df.index, unit="s")
    mock_backtest = {
        "base_balance": 1000,
        "exit_on_end": True,
        "comission": 0.00,
        "lot_size_perc": 1,
    }
    mock_df["action"] = ["e", "h", "x", "x", "x", "e", "x", "h", "h"]

    df = apply_logic_to_df(mock_df, mock_backtest)

    assert list(df.in_trade) == [
        True,
        True,
        False,
        False,
        False,
        True,
        False,
        False,
        False,
    ]

    assert list(df.account_value) == [
        0.0,
        0.0,
        2296.0,
        2296.0,
        2296.0,
        0.0,
        2274.32014388,
        2274.32014388,
        2274.32014388,
    ]

    assert list(df.adj_account_value) == [
        1000.0,
        1404.0,
        2296.0,
        2296.0,
        2296.0,
        2296.0,
        2274.32014388,
        2274.32014388,
        2274.32014388,
    ]

    assert df.fee.sum() == 0.0


def test_apply_logic_to_df_lot_size():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    mock_df.index = pd.to_datetime(mock_df.index, unit="s")

    mock_backtest = {
        "base_balance": 1000,
        "exit_on_end": True,
        "comission": 0.00,
        "lot_size_perc": 0.5,
    }
    mock_df["action"] = ["e", "h", "h", "x", "h", "h", "e", "h", "h"]

    df = apply_logic_to_df(mock_df, mock_backtest)

    assert list(df.in_trade) == [
        True,
        True,
        True,
        False,
        False,
        False,
        True,
        True,
        True,
        False,
    ]

    assert list(df.adj_account_value) == [
        1000.0,
        1202.0,
        1648.0,
        1560.0,
        1560.0,
        1560.0,
        1560.0,
        1555.39718566,
        1539.8184294100001,
        1539.8184294100001,
    ]



from fast_trade.archive import binance_api, coinbase_api
import datetime
from .update_kline import update_kline
from .db_helpers import get_local_assets
import pprint
import typing


def get_assets(exchange: str = "local") -> typing.List[str]:
    assets = []
    try:
        if exchange == "binanceus":
            assets.extend(binance_api.get_available_symbols())
        elif exchange == "binancecom":
            assets.extend(binance_api.get_available_symbols(tld="com"))
        elif exchange == "coinbase":
            assets.extend(coinbase_api.get_asset_ids())
        elif exchange == "local":
            assets.extend(get_local_assets())
        else:
            raise ValueError(f"Exchange {exchange} not supported")
        assets.sort()
    except Exception as e:
        print(f"Error getting assets for {exchange}: {e}")
        raise e

    pprint.pprint(assets)
    print(f"Found {len(assets)} assets for {exchange}")
    return assets


def download_asset(
    symbol: str,
    exchange: str,
    start: typing.Union[str, datetime.datetime] = datetime.datetime.now()
    - datetime.timedelta(days=30),
    end: typing.Union[str, datetime.datetime] = datetime.datetime.now(),
):
    """
    Download a single asset from the given exchange

    Args:
        symbol (str): The symbol to download
        exchange (str): The exchange to download from
        start (str | datetime.datetime): The start date to download from. Defaults to 30 days ago Must be in
        UTC timezone and isoformat.
        end (str | datetime.datetime): The end date to download to. Defaults to now. Must be in UTC timezone
        and isoformat.
    """
    # default to 30 days ago
    if not isinstance(start, datetime.datetime):
        start = datetime.datetime.fromisoformat(start)
    if not isinstance(end, datetime.datetime):
        end = datetime.datetime.fromisoformat(end)

    # set the timezone to utc
    start = start.replace(tzinfo=datetime.timezone.utc)
    end = end.replace(tzinfo=datetime.timezone.utc)
    # make sure the symbol exists

    if exchange == "binanceus":
        # make sure the symbol exists
        if symbol not in binance_api.get_available_symbols():
            raise ValueError(f"Symbol {symbol} not found on Binance US")
        db_path = update_kline(symbol, exchange, start, end)
    elif exchange == "binancecom":
        if symbol not in binance_api.get_available_symbols(tld="com"):
            raise ValueError(f"Symbol {symbol} not found on Binance COM")
        db_path = update_kline(symbol, exchange, start, end)
    elif exchange == "coinbase":
        if symbol not in coinbase_api.get_asset_ids():
            raise ValueError(f"Symbol {symbol} not found on Coinbase")
        db_path = update_kline(symbol, exchange, start, end)
    else:
        raise ValueError(f"Exchange {exchange} not supported")

    print(f"Downloaded {symbol} from {exchange} to {db_path}")
    return db_path



import pandas as pd
import os
import sqlite3
import typing
import datetime

ARCHIVE_PATH = os.getenv("ARCHIVE_PATH", os.path.join(os.getcwd(), "ft_archive"))


# update the kline archive by the given symbol and exchange
# get the archive path from the environment variable
def get_local_assets() -> typing.List[typing.Tuple[str, str]]:
    """
    Get the local assets from the archive

    Returns:
        typing.List[typing.Tuple[str, str]]: A list of tuples containing the exchange and symbol
    """
    all_assets = []

    for exchange in os.listdir(ARCHIVE_PATH):
        for symbol in os.listdir(os.path.join(ARCHIVE_PATH, exchange)):
            if symbol.startswith("_") or not symbol.endswith(".sqlite"):
                # ignore files that start with an underscore
                continue
            all_assets.append((exchange, symbol.replace(".sqlite", "")))

    return all_assets


def update_klines_to_db(df, symbol, exchange) -> str:
    """
    Store the kline dataframe to the db

    Args:
        df (pd.DataFrame): The kline dataframe to store
        symbol (str): The symbol of the klines
        exchange (str): The exchange of the klines

    Returns:
        str: The path to the db
    """
    # create the archive path if it doesn't exist
    if not os.path.exists(ARCHIVE_PATH):
        os.makedirs(ARCHIVE_PATH)
    # create the exchange path if it doesn't exist
    exchange_path = f"{ARCHIVE_PATH}/{exchange}"
    if not os.path.exists(exchange_path):
        os.makedirs(exchange_path)
    # create the symbol path if it doesn't exist
    symbol_path = f"{exchange_path}/{symbol}.sqlite"
    engine = connect_to_db(symbol_path, create=True)
    df = standardize_df(df)
    df.to_sql("klines", con=engine, if_exists="append", index=True, index_label="date")

    return symbol_path


def connect_to_db(db_path: str, create: bool = False) -> sqlite3.Connection:
    """
    Connect to the sqlite database

    Args:
        db_name (str, optional): The name of the database to connect to. Defaults to "ftc".

    Returns:
        sqlite3.Connection: The connection to the database
    """
    conn_str = db_path
    # check if the db exists
    if not os.path.exists(conn_str) and not create:
        raise Exception(f"Database {conn_str} does not exist")

    conn = sqlite3.connect(conn_str)
    # if db_name == "ftc":
    conn.execute("pragma journal_mode=WAL")
    return conn


def standardize_df(df):
    new_df = df.copy()

    new_df = new_df[~new_df.index.duplicated(keep="last")]

    # drop any columns that arent klines
    allowed_columns = [
        "open",
        "close",
        "high",
        "low",
        "volume",
    ]
    new_df = new_df[allowed_columns]
    new_df = new_df.sort_index()

    new_df.open = pd.to_numeric(new_df.open)
    new_df.close = pd.to_numeric(new_df.close)
    new_df.high = pd.to_numeric(new_df.high)
    new_df.low = pd.to_numeric(new_df.low)
    new_df.volume = pd.to_numeric(new_df.volume)

    return new_df


def get_kline(
    symbol: str,
    exchange: str,
    start_date: datetime.datetime = None,
    end_date: datetime.datetime = None,
    freq: str = "1Min",
) -> pd.DataFrame:
    """
    Get the klines from the db
    """
    db_path = f"{ARCHIVE_PATH}/{exchange}/{symbol}.sqlite"
    # if the db exists, if not try and downlaod it
    if not os.path.exists(db_path):
        import fast_trade.archive.update_kline as update_kline

        update_kline.update_kline(
            symbol=symbol, exchange=exchange, start_date=start_date, end_date=end_date
        )

    if start_date is not None:
        if isinstance(start_date, str):
            start_date = datetime.datetime.fromisoformat(start_date)

    if end_date is not None:
        if isinstance(end_date, str):
            end_date = datetime.datetime.fromisoformat(end_date)

    conn = connect_to_db(db_path)
    query = "SELECT * FROM klines"
    if start_date:
        query += f" WHERE date >= '{start_date.isoformat()}'"

    if end_date:
        query += f" AND date <= '{end_date.isoformat()}'"

    df = pd.read_sql_query(query, conn)
    df.date = pd.to_datetime(df.date)
    df = df.set_index("date")
    # set the freq of the dataframe
    df = df.resample(freq).agg(
        {
            "open": "first",
            "high": "max",
            "low": "min",
            "close": "last",
            "volume": "sum",
        }
    )

    return df


if __name__ == "__main__":
    symbol = "BTCUSDT"
    exchange = "binanceus"
    start_date = datetime.datetime(2024, 12, 12)
    end_date = datetime.datetime(2024, 12, 31)
    df = get_kline(symbol, exchange, start_date, end_date)
    print(df)



import math
import requests
import datetime
import time
import random
import pandas as pd
from pprint import pprint
import typing
import os

API_DELAY = os.getenv("API_DELAY", 0.3)
BASE_URL = "https://api.exchange.coinbase.com"
CB_REST_HEADER_MATCH = [
    "date",
    "low",
    "high",
    "open",
    "close",
    "volume",
]


def get_products() -> typing.List[dict]:
    """Returns a list of all tradable assets on Coinbase Pro"""
    try:
        res = requests.get(f"{BASE_URL}/products")
        if res.status_code > 399:
            print("unauthorized")
            return []
        return res.json()
    except Exception as e:
        print(e.status_code)
        print("Error fetching products: e")
        return []


def get_asset_ids() -> typing.List[str]:
    """Returns a list of all tradable assets on Coinbase Pro"""
    ids = [asset["id"] for asset in get_products()]
    ids.sort()
    return ids


def get_product_candles(
    product_id: str,
    start: datetime = None,
    end: datetime = None,
    update_status: callable = lambda x: None,
    store_func: callable = lambda x: None,
):
    """Returns the candle data for a given product"""

    # print("Fetching data for: ", product_id, start, end)
    if not start:
        # fetch the oldest date for this symbol
        start = get_oldest_day(product_id)

    # can't be more than 1 hour in the past
    end = end or datetime.datetime.utcnow()
    start = start or (end - datetime.timedelta(hours=3))

    # end = end + datetime.timedelta(days=1)

    start = start.replace(tzinfo=datetime.timezone.utc)
    end = end.replace(tzinfo=datetime.timezone.utc)

    # print("start", start, "end", end)
    # return
    # while datetime.datetime.fromisoformat(
    currentDate = start
    df = pd.DataFrame()

    # calculate the estimated number of calls
    total_duration_hours = (end - start).total_seconds() / 3600
    num_calls = math.ceil(total_duration_hours / 3)
    # print("estimated number of calls: ", num_calls)
    call_count = 0
    # print("start", start, "end", end)
    status_obj = {}
    bad_errors = 0
    start_time = time.time()
    while currentDate < end:
        # print("currentDate", currentDate, "end", end)
        currentDate = currentDate.replace(tzinfo=datetime.timezone.utc)
        next_end = currentDate + datetime.timedelta(hours=3)
        now = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)
        next_end = next_end.replace(tzinfo=datetime.timezone.utc)
        if next_end > now:
            next_end = now

        params = {
            "granularity": 60,
            "start": str(int(currentDate.timestamp())),
            "end": str(int(next_end.timestamp())),
        }

        df = get_single_candle(product_id, params, df)
        if df.empty:
            print(f"Error Downloading {product_id}")
            bad_errors += 1
            if bad_errors > 4:
                raise Exception(
                    f"Error Downloading: for {product_id} from {start} to {end}"
                )
            time.sleep(2 * bad_errors)
            continue
        call_count += 1
        if call_count % 10 == 0:

            store_func(df, product_id, "coinbase")

        status_obj = {
            "symbol": product_id,
            "perc_complete": round(call_count / num_calls * 100, 2),
            "call_count": call_count,
            "total_calls": num_calls,
            "total_time": round(time.time() - start_time, 2),
            # "sleep_time": sleeper,
            "est_time_remaining": round(
                (time.time() - start_time) / call_count * (num_calls - call_count), 2
            ),
        }
        update_status(status_obj)
        currentDate = next_end
    if not df.empty:
        df.sort_index(inplace=True, ascending=False)

    return df, status_obj


def get_single_candle(product_id: str, params: dict = {}, df=pd.DataFrame()):
    url = f"{BASE_URL}/products/{product_id}/candles"
    headers = {"Content-Type": "application/json"}
    try:
        res = requests.get(url, params=params, headers=headers)
        bad_errors = 0
        if res.status_code > 399:
            bad_errors += 1
            # skip this call
            print("Error ", res.status_code, res.text)
            # bad_errors += 1
            if bad_errors > 1:
                time.sleep(bad_errors * bad_errors * bad_errors)

            if bad_errors > 5:
                raise Exception(f"Api Error: {res.status_code} {res.text}")
            # Warning("skipping call ", res.status_code, res.text)
        # print("res", res.status_code, res.text)
        res = res.json()
        new_df = df_from_candles(res)
        if new_df.empty:
            bad_errors += 1
            if bad_errors > 1:
                raise Exception(f"Error Downloading: for {product_id}")
            return pd.DataFrame()
        sleep_time = random.random() * 0.5 + 0.1
        time.sleep(sleep_time)
        df = pd.concat([df, new_df])
        df.drop_duplicates(inplace=True)
        return df
    except Exception as e:
        print("error", e)
        return pd.DataFrame()


def df_from_candles(klines):
    new_df = pd.DataFrame(klines, columns=CB_REST_HEADER_MATCH)
    new_df = new_df.drop_duplicates()
    new_df.index = pd.to_datetime(new_df.date, unit="s")

    if "date" in new_df.columns:
        new_df = new_df.drop(columns=["date"])

    return new_df


def get_oldest_day(
    product_id, start_date=datetime.datetime(2015, 7, 21)
) -> datetime.datetime:
    """Find the oldest day with data for a given product_id."""

    url = f"{BASE_URL}/products/{product_id}/candles"
    end_date = datetime.datetime.now()
    call_count = 0
    while start_date <= end_date:
        middle_date = start_date + (end_date - start_date) // 2
        params = {
            "granularity": 60,
            "start": int(middle_date.timestamp()),
            "end": int((middle_date + datetime.timedelta(minutes=1)).timestamp()),
        }

        response = requests.get(url, params=params)
        call_count += 1
        time.sleep(random.random() * 0.2 + 0.1)
        if response.status_code != 200:
            time.sleep(2)
            raise Exception(f"API request failed: {response.text}")

        data = response.json()
        if data:
            # Data found, search in earlier half
            end_date = middle_date - datetime.timedelta(days=1)
        else:
            # No data, search in later half
            start_date = middle_date + datetime.timedelta(days=1)

    print(f"API calls: {call_count}")
    return end_date


if __name__ == "__main__":
    start = datetime.datetime(2024, 2, 7)
    res = get_product_candles("BTC-USD", start=start)
    # res = find_oldest_date("DOGE-USD")
    # res = get_products()
    pprint(res)

    # res.to_csv("btc.csv")
    # print(res)
    # print(res[0], res[1])



from datetime import datetime
import pandas as pd
import os
import re

from .transformers_map import transformers_map


def build_data_frame(backtest: dict, csv_path: str):
    """Creates a Pandas DataFame with the provided backtest. Used when providing a CSV as the datafile

    Parameters
    ----------
    backtest: dict, provides instructions on how to build the dataframe
    csv_path: string, absolute path of where to find the data file

    Returns
    -------
    object, A Pandas DataFrame indexed buy date
    """
    df = load_basic_df_from_csv(csv_path)

    if df.empty:
        raise Exception("Dataframe is empty. Check the start and end dates")

    df = prepare_df(df, backtest)

    return df


def load_basic_df_from_csv(csv_path: str):
    """Loads a dataframe from a csv
    Parameters
    ----------
        csv_path: string, path to the csv so it can be read

    Returns
        df, A basic dataframe with the data from the csv
    """

    if not os.path.isfile(csv_path):
        raise Exception(f"File not found: {csv_path}")

    df = pd.read_csv(csv_path, header=0)
    df = standardize_df(df)

    return df


def prepare_df(df: pd.DataFrame, backtest: dict):
    """Prepares the provided dataframe for a backtest by applying the datapoints and splicing based on the given backtest.
        Useful when loading an existing dataframe (ex. from a cache).

    Parameters
    ----------
        df: DataFrame, should have all the open, high, low, close, volume data set as headers and indexed by date
        backtest: dict, provides instructions on how to build the dataframe

    Returns
    ------
        df: DataFrame, with all the datapoints as column headers and trimmed to the provided time frames
    """

    datapoints = backtest.get("datapoints", [])

    if backtest.get("chart_period"):
        # raise a warning if chart_period is not a valid frequency
        if not infer_frequency(df, backtest.get("chart_period")):
            raise ValueError(f"Invalid chart period: {backtest.get('chart_period')}")
        freq = backtest.get("chart_period")
    else:
        freq = backtest.get("freq", "1Min")

    start_time = backtest.get("start")
    stop_time = backtest.get("stop")
    df = apply_charting_to_df(df, freq, start_time, stop_time)
    df = apply_transformers_to_dataframe(df, datapoints)
    trailing_stop_loss = backtest.get("trailing_stop_loss", 0)
    if trailing_stop_loss:
        df["trailing_stop_loss"] = df["close"].cummax() * (
            1 - float(trailing_stop_loss)
        )

    return df


def apply_charting_to_df(df: pd.DataFrame, freq: str, start_time: str, stop_time: str):
    """Modifies the dataframe based on the freq, start dates and end dates
    Parameters
    ----------
        df: dataframe with data loaded
        freq: string, describes how often to sample data, default is '1Min' (1 minute)
            see https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
        start_time: datestring in YYYY-MM-DD HH:MM (ex. 2020-08-31 04:00) of when to begin the backtest
        stop_time: datestring of YYYY-MM-DD HH:MM when to stop the backtest
    Returns
        DataFrame, a sorted dataframe ready for consumption by run_backtest
    """
    if df.index.dtype != "datetime64[ns]":
        headers = df.columns.values.tolist()
        headers.extend([df.index.name])
        if "date" not in headers:
            raise Exception(
                "Data does not have a date column. Headers must include date, open, high, low, close, volume."
            )

        time_unit = detect_time_unit(df.date[1])
        df.date = pd.to_datetime(df.date, unit=time_unit)
        df.set_index("date", inplace=True)
    if start_time:
        if isinstance(start_time, datetime) or type(start_time) is int:
            time_unit = detect_time_unit(start_time)
            start_time = pd.to_datetime(start_time, unit=time_unit)
            start_time = start_time.strftime("%Y-%m-%d %H:%M:%S")

    if stop_time:
        if isinstance(stop_time, datetime) or type(stop_time) is int:
            time_unit = detect_time_unit(stop_time)
            stop_time = pd.to_datetime(stop_time, unit=time_unit)
            stop_time = stop_time.strftime("%Y-%m-%d %H:%M:%S")

    df = df.resample(freq).first()

    if start_time and stop_time:
        df = df[start_time:stop_time]  # noqa
    elif start_time and not stop_time:
        df = df[start_time:]  # noqa
    elif not start_time and stop_time:
        df = df[:stop_time]

    return df


def apply_transformers_to_dataframe(
    df: pd.DataFrame,
    transformers: list,
):
    """Applies indications from the backtest to the dataframe
    Parameters
    ----------
        df: dataframe loaded with data
        transformers: list of transformers as dictionary objects

        transformer detail:
        {
            "transformer": "", string, actual function to be called MUST be in the transformers_map
            "name": "", string, name of the transformer, becomes a column on the dataframe
            "args": [], list arguments to pass the the function,
            "freq": "", string, frequency of the transformer, default is the freq in the backtest
        }

    Returns
    -------
        df, a modified dataframe with all the datapoints calculated as columns
    """

    base_freq = infer_frequency(df)
    # set the freq of the dataframe
    df = df.asfreq(base_freq)
    # return df
    for ind in transformers:
        transformer = ind.get("transformer")
        field_name = ind.get("name")
        freq = ind.get("freq", None)

        # Create a temporary dataframe with the desired frequency
        if freq:
            tmp_df = (
                df.resample(freq)
                .agg(
                    {
                        "open": "first",
                        "high": "max",
                        "low": "min",
                        "close": "last",
                        "volume": "sum",
                    }
                )
                .ffill()
            )
        else:
            tmp_df = df

        if len(ind.get("args", [])):
            args = ind.get("args")
            trans_res = transformers_map[transformer](tmp_df, *args)
        else:
            trans_res = transformers_map[transformer](tmp_df)

        if isinstance(trans_res, pd.DataFrame):
            df = process_res_df(df, ind, trans_res)
            # resample the dataframe to the base freq

        elif isinstance(trans_res, pd.Series):
            df[field_name] = trans_res

        df = df.asfreq(base_freq).ffill()

    return df


def process_res_df(df, ind, trans_res):
    """handle if a transformer returns multiple columns
    To manage this, we just add the name of column in a clean
    way, removing periods and lowercasing it.

    Parameters
    ----------
    df, dataframe, current dataframe
    ind, indicator object
    trans_res, result from the transformer function

    Returns
    -------
    df, dataframe, updated dataframe with the new columns
    """
    for key in trans_res.keys().values:
        i_name = ind.get("name")
        clean_key = key.lower()
        clean_key = clean_key.replace(".", "")
        clean_key = clean_key.replace(" ", "_")
        df_key = f"{i_name}_{clean_key}"
        df[df_key] = trans_res[key]

    return df


def detect_time_unit(str_or_int: str or int):
    """Determines a if a timestamp is really a timestamp and if it
    matches is in seconds or milliseconds
    Parameters
    ----------
        str_or_int: string or int of the timestamp to detect against

    Returns
    -------
        string of "s" or "ms", or None if nothing detected
    """
    str_or_int = str(str_or_int)
    regex1 = r"^(\d{10})$"
    regex2 = r"^(\d{13})$"

    if re.match(regex1, str_or_int):
        return "s"
    if re.match(regex2, str_or_int):
        return "ms"


def standardize_df(df: pd.DataFrame):
    """Standardizes a dataframe with the basic features used
    throughout the project.
    Parameters
    ----------
        df: A pandas dataframe (probably one just created) with
    at least the required columns of: date, open, close, high, low, volume.

    Returns
    -------
        A new pandas dataframe of with all the data in the expected types.
    """
    new_df = df.copy()

    if "date" in new_df.columns:
        new_df = new_df.set_index("date")

    ts = str(new_df.index[0])

    time_unit = detect_time_unit(ts)

    new_df.index = pd.to_datetime(new_df.index, unit=time_unit)
    new_df = new_df[~new_df.index.duplicated(keep="first")]
    new_df = new_df.sort_index()

    columns_to_drop = ["ignore", "date"]

    new_df.drop(columns=columns_to_drop, errors="ignore")

    new_df.open = pd.to_numeric(new_df.open)
    new_df.close = pd.to_numeric(new_df.close)
    new_df.high = pd.to_numeric(new_df.high)
    new_df.low = pd.to_numeric(new_df.low)
    new_df.volume = pd.to_numeric(new_df.volume)

    return new_df


def infer_frequency(df: pd.DataFrame) -> str:
    """Infers the frequency of a DataFrame by analyzing time differences between consecutive index values.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with a datetime index

    Returns
    -------
    str
        The inferred frequency as a string (e.g., '1Min', '5Min', '1H', '1D', etc.)
        Returns None if frequency cannot be determined
    """
    if not isinstance(df.index, pd.DatetimeIndex):
        raise ValueError("DataFrame index must be a DatetimeIndex")

    if df.index.freq is not None:
        return df.index.freq

    # Calculate time differences between consecutive index values
    time_diffs = df.index.to_series().diff()

    # Get the most common time difference
    most_common_diff = time_diffs.mode()[0]
    seconds = most_common_diff.total_seconds()

    # Convert seconds to appropriate frequency string
    if seconds < 60:
        return f"{int(seconds)}S"
    elif seconds < 3600:
        minutes = int(seconds / 60)
        return f"{minutes}Min"
    elif seconds < 86400:
        hours = int(seconds / 3600)
        return f"{hours}H"
    else:
        days = int(seconds / 86400)
        return f"{days}D"



import datetime
import os
from .binance_api import get_binance_klines
from .coinbase_api import get_product_candles
from .db_helpers import update_klines_to_db
import pandas as pd
import typing

supported_exchanges = ["binanceus", "binancecom", "coinbase"]


def update_kline(
    symbol,
    exchange,
    start_date: typing.Optional[datetime.datetime] = None,
    end_date: typing.Optional[datetime.datetime] = None,
):
    if exchange not in supported_exchanges:
        raise ValueError(f"Exchange {exchange} not supported")

    if start_date is None:
        # 30 days ago
        start_date = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
            days=30
        )
    if end_date is None:
        # now
        end_date = datetime.datetime.now(datetime.timezone.utc)

    start_date = start_date.replace(tzinfo=datetime.timezone.utc)
    end_date = end_date.replace(tzinfo=datetime.timezone.utc)
    now = datetime.datetime.now(datetime.timezone.utc)

    if end_date > now:
        end_date = now.replace(second=0, microsecond=0)

    klines = pd.DataFrame()
    curr_date = start_date

    def status_update(status_obj):
        # os.system("clear")
        # print(status_obj)
        in_seconds = status_obj["est_time_remaining"]
        in_minutes = in_seconds / 60
        msg = f"Downloading {symbol} from {exchange}."
        if in_seconds > 100:
            msg += f" Remaining (in minutes): {in_minutes:.2f}"
        else:
            msg += f" Remaining (in seconds): {in_seconds:.2f}"

        msg += f" {status_obj['perc_complete']}% complete"
        # update the db
        print(msg)

    if exchange == "binanceus":
        klines, status_obj = get_binance_klines(
            symbol,
            curr_date,
            end_date,
            "us",
            status_update,
            store_func=update_klines_to_db,
        )
    elif exchange == "binancecom":
        klines, status_obj = get_binance_klines(
            symbol,
            curr_date,
            end_date,
            "com",
            status_update,
            store_func=update_klines_to_db,
        )
    elif exchange == "coinbase":
        klines, status_obj = get_product_candles(
            symbol, curr_date, end_date, status_update, store_func=update_klines_to_db
        )
    else:
        raise ValueError(f"Exchange {exchange} not supported")

    db_path = update_klines_to_db(klines, symbol, exchange)

    return db_path


if __name__ == "__main__":
    # symbol = "BTCUSDT"
    # exchange = "binanceus"
    symbol = "BTC-USD"
    exchange = "coinbase"
    start_date = datetime.datetime(2024, 1, 1)
    end_date = datetime.datetime(2024, 1, 14)
    res = update_kline(symbol, exchange, start_date, end_date)
    print(res)



import re

import pandas as pd

from .transformers_map import transformers_map


def validate_backtest(backtest):
    """validates a backtest object and returns errors/warnings

    Parameters
    -----------
        backtest, user constructed backtest object

    Returns
    -------
        tuple, [errors, warnings], returns a tuple with any
        errors and warnings, respectively
    """

    backtest_mirror = {
        "base_balance": None,
        "freq": None,
        "chart_start": None,
        "chart_stop": None,
        "comission": None,
        "datapoints": None,
        "enter": None,
        "exit": None,
        "any_enter": None,
        "any_exit": None,
        "trailing_stop_loss": None,
        "exit_on_end": None,
    }

    required_keys = [
        "datapoints",
        "enter",
        "exit",
        "start_date",
    ]

    curr_keys = list(backtest.keys())

    for req in required_keys:
        if req not in curr_keys:
            backtest_mirror[req] = {
                "error": True,
                "msgs": [f'Paramater "{req}" required'],
            }

    base_balance = backtest.get("base_balance")
    if base_balance:
        bb = match_field_type_to_value(base_balance)

        if isinstance(bb, str):
            backtest_mirror["base_balance"] = {
                "error": True,
                "msgs": ["base_balance must be a float or string"],
            }
    freq = backtest.get("freq")
    if freq:
        if not re.search(r"(^\d{1,4}((T)|(Min)|(H)|(h)|(D)|)$)", freq):
            backtest_mirror["freq"] = {
                "error": True,
                "msgs": ["Chart period not valid"],
            }

    # check the datapints
    dp_errors = []

    for dp in backtest.get("datapoints", []):
        # check that we have the transformer function
        trans = dp.get("transformer", "")
        if trans not in list(transformers_map.keys()):
            dp_errors.append(
                f'Tranformer "{trans}" not valid. See the transformers_map for valid transformers.'
            )

        if not dp.get("name"):
            dp_errors.append("Name is required.")

    if len(dp_errors):
        backtest_mirror["datapoints"] = {"error": True, "msgs": dp_errors}

    # check the logics

    # fill the indicator keys
    basic_keys = ["open", "high", "low", "close", "volume"]
    indicator_keys = [
        dp.get("name") for dp in backtest.get("datapoints", []) if dp.get("name")
    ]
    indicator_keys.extend(basic_keys)

    exit_errors = []
    any_exit_errors = []
    enter_errors = []
    any_enter_errors = []
    # by default these can always be used
    indicator_keys = list(set(indicator_keys))
    allowed_operators = [">", "=", "<", ">=", "<="]

    for logic in backtest.get("exit", []):
        for log in logic:
            if (
                log not in indicator_keys
                and isinstance(log, str)
                and log not in allowed_operators
                and not log.isnumeric()
            ):
                exit_errors.append(
                    f'Operator "{log}" referenced in exit logic not found in datapoints. Check datapoints and logic.'
                )

        if len(exit_errors):
            backtest_mirror["exit"] = {"error": True, "msgs": exit_errors}

    for logic in backtest.get("enter", []):
        for log in logic:
            if (
                log not in indicator_keys
                and isinstance(log, str)
                and log not in allowed_operators
                and not log.isnumeric()
            ):
                enter_errors.append(
                    f'Datapoint "{log}" referenced in enter logic not found in datapoints. Check datapoints and logic.'
                )

        if len(enter_errors):
            backtest_mirror["enter"] = {"error": True, "msgs": enter_errors}

    for logic in backtest.get("any_enter", []):
        for log in logic:
            if (
                log not in indicator_keys
                and isinstance(log, str)
                and log not in allowed_operators
            ):
                any_enter_errors.append(
                    f'Datapoint "{log}" referenced in any_enter \
                    logic not found in datapoints. Check datapoints and logic.'
                )
        if len(any_enter_errors):
            backtest_mirror["any_enter"] = {"error": True, "msgs": any_enter_errors}

    for logic in backtest.get("any_exit", []):
        for log in logic:
            if (
                log not in indicator_keys
                and isinstance(log, str)
                and log not in allowed_operators
            ):
                any_exit_errors.append(
                    f'Datapoint "{log}" referenced in any_exit logic not found in datapoints. Check datapoints and logic.'
                )

        if len(any_exit_errors):
            backtest_mirror["any_exit"] = {"error": True, "msgs": any_exit_errors}

    lot_size = backtest.get("lot_size", 0)

    if lot_size > 1:
        backtest_mirror["lot_size"] = {
            "error": True,
            "msgs": ["Lot size must be less than 1."],
        }

    if lot_size < 0:
        backtest_mirror["lot_size"] = {
            "error": True,
            "msgs": ["Lot size be larger than 0."],
        }

    return_value = backtest_mirror
    return_value["has_error"] = any(
        [backtest_mirror[key] for key in backtest_mirror.keys()]
    )
    return return_value


def match_field_type_to_value(field):
    if isinstance(field, str):
        if field.isnumeric():
            return int(field)
        if re.match(r"^-?\d+(?:\.\d+)$", field):  # if its a string in a float
            return float(field)
    return field


def validate_backtest_with_df(backtest: dict, df: pd.DataFrame) -> None:
    errors = validate_backtest(backtest)
    if errors.get("has_error"):
        raise Exception(errors)

    if df.empty:
        raise Exception("Dataframe is empty. Check your data source.")

    errors = []

    for dp in backtest.get("datapoints", []):
        if dp.get("name") not in df.columns:
            errors.append(f'Datapoint "{dp.get("name")}" not found in dataframe.')
    if len(errors):
        raise Exception(errors)



import datetime
import numpy as np
import pandas as pd
import pytest
from fast_trade.calculate_perc_missing import calculate_perc_missing
import random


def test_calculate_perc_missing_none_missing():
    # generate a list of dates from the first to the last date in the dataframe
    today = datetime.datetime.now()
    last_week = today - datetime.timedelta(days=7)
    mock_dates = pd.date_range(start=last_week, end=today, freq="1Min")
    mock_df = pd.DataFrame(index=mock_dates)
    mock_df["close"] = 0
    mock_df = mock_df.asfreq("1Min")
    [perc_missinng, total_missing] = calculate_perc_missing(mock_df)

    assert perc_missinng == 0.0
    assert total_missing == 0.0


def test_calculate_perc_missing_some_missing():
    today = datetime.datetime.now().replace(second=0, microsecond=0, hour=12, minute=0)
    last_week = today - datetime.timedelta(hours=1)
    mock_dates = pd.date_range(start=last_week, freq="1Min", periods=10)
    mock_df = pd.DataFrame(index=mock_dates)
    mock_df = mock_df.asfreq("1Min")

    # remove 2 dates from the list
    mock_df = mock_df.drop(mock_df.index[-5])
    mock_df = mock_df.drop(mock_df.index[-5])
    mock_df["close"] = 0

    [perc_missinng, total_missing] = calculate_perc_missing(mock_df)

    assert perc_missinng == 20
    assert total_missing == 2


def test_calculate_perc_missing_empty_df():
    mock_df = pd.DataFrame()
    with pytest.raises(ValueError):
        calculate_perc_missing(mock_df)


def test_calculate_perc_missing_no_index():
    mock_df = pd.DataFrame({"close": [1, 2, 3, 4, 5]})
    with pytest.raises(ValueError):
        calculate_perc_missing(mock_df)


def test_calculate_perc_missing_some_with_different_freq():
    today = datetime.datetime.now().replace(second=0, microsecond=0, hour=12, minute=0)
    last_week = today - datetime.timedelta(hours=1)
    # freq = random.choice(["1Min", "5Min", "15Min", "30Min", "1H", "4H", "1D"])
    mock_dates = pd.date_range(start=last_week, freq="10Min", periods=10)
    # remove 2 dates from the list
    # mock_dates = mock_dates.drop(mock_dates[-5])
    # mock_dates = mock_dates.drop(mock_dates[-5])
    # mock_dates = mock_dates.drop(mock_dates[-5])
    mock_df = pd.DataFrame(index=mock_dates)
    mock_df = mock_df.asfreq("1Min")

    # remove 2 dates from the list
    mock_df = mock_df.drop(mock_df.index[-5])
    mock_df = mock_df.drop(mock_df.index[-5])
    mock_df["close"] = 0

    [perc_missinng, total_missing] = calculate_perc_missing(mock_df)

    assert perc_missinng == 2.2
    assert total_missing == 2



def handle_rule(result: dict, rule: list) -> bool:
    """
    Handle a single rule with support for dotted notation in nested dictionaries.
    Example: Using "a.b.c" will access result["a"]["b"]["c"]

    Returns
    -------
    bool, the result of the rule
    """
    column_name = rule[0]
    operator = rule[1]
    value_or_column_name = rule[2]

    # Handle nested dictionary access using dotted notation
    def get_nested_value(d: dict, key: str):
        keys = key.split(".")
        value = d
        for k in keys:
            value = value[k]
        return value

    # Get the actual value for comparison
    result_value = float(get_nested_value(result, column_name))

    # Handle the comparison value
    if isinstance(value_or_column_name, str):
        value = float(get_nested_value(result, value_or_column_name))
    else:
        value = float(value_or_column_name)

    if operator == ">":
        return result_value > value
    elif operator == "<":
        return result_value < value
    elif operator == ">=":
        return result_value >= value
    elif operator == "<=":
        return result_value <= value
    return False


def evaluate_rules(result: dict, rules: list) -> tuple:
    """
    Evaluate a backtest result against a list of rules. Useful for quickly checking if a backtest result meets certain criteria.

    Parameters
    ----------
    result: dict, the result of the backtest
    rules: list, the rules to evaluate

    Rule:
    [
        "column_name", # column name to compare
        "operator", # operator to compare
        "value_or_column_name" # value to compare
    ]

    Returns
    -------
    tuple, (all(res), any(res), res)
    """
    res = []
    for rule in rules:
        try:
            res.append(handle_rule(result, rule))
        except Exception as e:
            print(e)

    if len(res) == 0:
        return False, False, []

    return all(res), any(res), res


if __name__ == "__main__":
    rules = [["trade_streaks.avg_win_streak", ">", 6]]
    result = {"trade_streaks": {"avg_win_streak": 5}}
    print(evaluate_rules(result, rules))



import pytest
import pandas as pd
import datetime

from fast_trade.build_data_frame import (
    build_data_frame,
    detect_time_unit,
    load_basic_df_from_csv,
    apply_transformers_to_dataframe,
    apply_charting_to_df,
    prepare_df,
    process_res_df,
)


def test_detect_time_unit_s():
    mock_timestring = 1595115901734
    result = detect_time_unit(mock_timestring)

    assert result == "ms"


def test_detect_time_unit_ms():
    mock_timestring = 1595115901
    result = detect_time_unit(mock_timestring)

    assert result == "s"


def test_load_basic_df_from_csv_str_1():
    mock_data_path = "./test/ohlcv_data.csv.txt"

    result_df = load_basic_df_from_csv(mock_data_path)
    header = list(result_df.head())
    assert "close" in header
    assert "open" in header
    assert "high" in header
    assert "low" in header
    assert "volume" in header

    assert result_df.index.name == "date"


def test_load_basic_df_from_csv_list_1():
    mock_data_path = "./test/ohlcv_data.csv.txt"

    result_df = load_basic_df_from_csv(mock_data_path)

    expected_line = [0.01404, 0.01, 0.025, 0.01, 3117.0]

    assert list(result_df.iloc[1]) == expected_line


def test_load_basic_df_from_csv_str_error_1():
    mock_data_path = "./test/SomeFakeNews.csv.txt"

    with pytest.raises(Exception, match=r"File not found:*"):
        load_basic_df_from_csv(mock_data_path)


def test_apply_transformers_to_dataframe_1_ind():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_transformers = [
        {"transformer": "sma", "name": "example_transformer_name", "args": [3]}
    ]
    mock_df.index = pd.to_datetime(mock_df.date, unit="s")

    result_df = apply_transformers_to_dataframe(mock_df, mock_transformers)

    header = list(result_df.head())

    assert "example_transformer_name" in header
    assert "FAKE_transformer_name" not in header


def test_apply_transformers_to_dataframe_no_args():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_df.index = pd.to_datetime(mock_df.date, unit="s")
    mock_transformers = [{"transformer": "rsi", "name": "rsi", "args": []}]

    result_df = apply_transformers_to_dataframe(mock_df, mock_transformers)

    assert "rsi" in list(result_df.columns)


def test_apply_transformers_to_dataframe_no_args_multi_col():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_df.index = pd.to_datetime(mock_df.date, unit="s")
    mock_transformers = [{"transformer": "wto", "name": "wto", "args": []}]

    result_df = apply_transformers_to_dataframe(mock_df, mock_transformers)

    assert "wto_wt1" in list(result_df.columns)
    assert "wto_wt2" in list(result_df.columns)


def test_apply_transformers_to_dataframe_freq():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_transformers = [
        {
            "transformer": "sma",
            "name": "example_transformer_name",
            "args": [3],
            "freq": "3Min",
        }
    ]

    mock_df.index = pd.to_datetime(mock_df.date, unit="s")

    result_df = apply_transformers_to_dataframe(mock_df, mock_transformers)

    assert "example_transformer_name" in list(result_df.columns)
    assert result_df.index.freq == "1Min"


def test_apply_charting_to_df_1():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", index_col="date")
    # mock_df.set_index(["date"], inplace=True)
    mock_df.index = pd.to_datetime(mock_df.index, unit="s")
    mock_freq = "2Min"
    mock_start_time = "2018-04-17"
    mock_stop_time = ""

    result_df = apply_charting_to_df(
        mock_df, mock_freq, mock_start_time, mock_stop_time
    )

    assert (result_df.iloc[2].name - result_df.iloc[1].name).total_seconds() == 120
    assert (result_df.iloc[4].name - result_df.iloc[1].name).total_seconds() == 360


def test_apply_charting_to_df_2():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_df.set_index(["date"], inplace=True)
    mock_df.index = pd.to_datetime(mock_df.index, unit="s")
    mock_freq = "1Min"
    mock_start_time = "2018-04-17 04:00:00"
    mock_stop_time = "2018-04-17 04:10:00"

    past_stop_time = datetime.datetime.strptime(
        "2018-04-17 04:11:00", "%Y-%m-%d %H:%M:%S"
    )
    result_df = apply_charting_to_df(
        mock_df, mock_freq, mock_start_time, mock_stop_time
    )

    assert result_df.iloc[-1].name < past_stop_time


def test_apply_charting_to_df_3():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_df.set_index(["date"], inplace=True)
    mock_df.index = pd.to_datetime(mock_df.index, unit="s")
    mock_freq = "1Min"
    mock_start_time = ""
    mock_stop_time = "2018-04-17 04:10:00"

    # past_stop_time = "2018-04-17 04:11:00"

    # result_df = apply_charting_to_df(
    #     mock_df, mock_freq, mock_start_time, mock_stop_time
    # )

    past_stop_time = datetime.datetime.strptime(
        "2018-04-17 04:11:00", "%Y-%m-%d %H:%M:%S"
    )
    result_df = apply_charting_to_df(
        mock_df, mock_freq, mock_start_time, mock_stop_time
    )

    assert result_df.index[0] < past_stop_time


def test_apply_charting_to_df_stop_time_int():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_df.index = pd.to_datetime(mock_df.date, unit="s")
    mock_freq = "1Min"
    mock_start_time = ""
    mock_stop_time = 1523938200

    past_stop_time = datetime.datetime.strptime(
        "2018-04-17 04:11:00", "%Y-%m-%d %H:%M:%S"
    )

    result_df = apply_charting_to_df(
        mock_df, mock_freq, mock_start_time, mock_stop_time
    )

    assert result_df.index[0] < past_stop_time


def test_apply_charting_to_df_start_time_int():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt")
    mock_df.index = pd.to_datetime(mock_df.date, unit="s")
    mock_freq = "1Min"
    mock_start_time = 1523938200
    mock_stop_time = ""

    past_stop_time = datetime.datetime.strptime(
        "2018-04-17 04:11:00", "%Y-%m-%d %H:%M:%S"
    )

    result_df = apply_charting_to_df(
        mock_df, mock_freq, mock_start_time, mock_stop_time
    )

    assert result_df.index[0] < past_stop_time


def test_process_res_df():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True)
    mock_df.index = pd.to_datetime(mock_df.date, unit="s")
    mock_ind = {"name": "ind_1"}
    val1 = [0, 1, 2, 3, 4, 5, 6, 7, 8]
    val2 = [8, 7, 6, 5, 4, 3, 2, 1, 0]
    mock_trans_res = pd.DataFrame(
        data={"Val 1": val1, "Val 2": val2},
        index=mock_df.index,
    )

    res = process_res_df(mock_df, mock_ind, mock_trans_res)

    assert list(res.ind_1_val_1.values) == val1
    assert list(res.ind_1_val_2.values) == val2


def test_apply_transformers_to_dataframe():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True)
    mock_df.index = pd.to_datetime(mock_df.date, unit="s")


def test_prepare_df():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True)
    mock_df.index = pd.to_datetime(mock_df.date, unit="s")

    mock_backtest = {
        "freq": "1Min",
        "start": "",
        "stop": "",
        "trailing_stop_loss": 0.1,
        "datapoints": [{"name": "ind_1", "args": [], "transformer": "ema"}],
    }

    res = prepare_df(mock_df, mock_backtest)

    assert "ind_1" in list(res.columns)
    assert "trailing_stop_loss" in list(res.columns)


def test_build_data_frame():
    mock_backtest = {
        "freq": "1Min",
        "start": "",
        "stop": "",
        "trailing_stop_loss": 0.01,
        "datapoints": [{"name": "indy_2", "args": [2], "transformer": "sma"}],
    }

    mock_csv_path = "./test/ohlcv_data.csv.txt"

    res = build_data_frame(mock_backtest, mock_csv_path)

    print(res)

    assert type(res) == pd.DataFrame


def test_build_data_frame_no_data():
    mock_backtest = {
        "freq": "1Min",
        "start": "",
        "stop": "",
        "trailing_stop_loss": 0.01,
        "datapoints": [{"name": "indy_2", "args": [2], "transformer": "sma"}],
    }

    mock_csv_path = "./test/empty.csv.txt"
    # with pytest.raises(RuntimeError) as excinfo:

    #     def f():
    #         f()

    #     f()
    # assert "maximum recursion" in str(excinfo.value)

    with pytest.raises(Exception) as exeinfo:
        build_data_frame(mock_backtest, mock_csv_path)

        assert "Dataframe is empty. Check the start and end dates" in str(exeinfo.value)



import datetime
import time
from .db_helpers import connect_to_db
import os
from .update_kline import update_kline


ARCHIVE_PATH = os.getenv("ARCHIVE_PATH", os.path.join(os.getcwd(), "ft_archive"))


def update_single_archive(symbol: str, exchange: str):
    # check the oldest date in the existing archive
    if not symbol.endswith(".sqlite"):
        symbol = symbol + ".sqlite"
    path = os.path.join(ARCHIVE_PATH, exchange, symbol)
    db = connect_to_db(path)

    now = datetime.datetime.now(datetime.timezone.utc)
    now = now.replace(second=0, microsecond=0)

    start_date = db.execute("SELECT max(date) FROM klines").fetchone()[0]

    if start_date is None:
        start_date = now - datetime.timedelta(days=7)
    else:
        start_date = datetime.datetime.fromisoformat(start_date)

    actual_symbol = symbol.replace(".sqlite", "")
    print(f"Updating {actual_symbol} from {exchange} from {start_date} to {now}")

    update_kline(
        symbol=actual_symbol,
        exchange=exchange,
        start_date=start_date,
        end_date=now,
    )


def update_archive():
    """Read the archive and update the klines"""
    count = 0
    start_time = time.time()

    for exchange in os.listdir(ARCHIVE_PATH):
        for symbol in os.listdir(os.path.join(ARCHIVE_PATH, exchange)):
            if symbol.startswith("_") or not symbol.endswith(".sqlite"):
                # ignore files that start with an underscore
                continue
            try:
                update_single_archive(symbol, exchange)
                count += 1
            except Exception as e:
                print(f"Error updating {symbol} from {exchange}: {e}")
                raise e

    updated_time = round(time.time() - start_time, 2)
    print(f"\n\nUpdated {count} symbols in {updated_time} seconds ✅")


if __name__ == "__main__":
    update_archive()
